% Sample code of the paper:
% 
% Jiahao Pang and Gene Cheung, "Graph Laplacian regularization for image 
% denoising: analysis in the continuous domain," IEEE Trans. Image Process., 
% vol. 26, no. 4, pp. 1770-1785, April 2017.
% 
% MATLAB code prepard by Jiahao Pang
% E-mail: jpang@connect.ust.hk or pangjiahao@sensetime.com
% 
% This script implements our denoising algorithm, optimal graph Laplacian
% regularization (OGLR) for image denoising

function img_den = oglr(img_noi, sigma_I, type)

%% initialization
addpath('nth_element_0.87');
[height, width] = size(img_noi);
img_int = img_noi;

%% noise-independent paramaters
search_rad     = 25;        % search radius
edge_min       = 4;         % minimum degree of the graphs, may use 4 or 8
iter_max       = 10;        % maximum number of iterations of denoising
reg_iter_max   = 30;        % maximum number of iterations of the Muller's method
lambda_0_const = 1e-6;
lambda_1_const = 5e1;       % control the search of lambda
c_sigma_I      = 0.05;      % control sigma_I
c_eps          = 0.04;      % control eps for constructing graphs
c_t_dct        = 1.5;       % determine the threshold of similar patches
alpha          = 1e-12;     % alpha in the paper
sigma_p        = 1e6;       % sigma_g in the paper
sigma_th       = 2;         % sigma_th in the paper, threshold for stop denoising
if strcmp(type, 'ntl') == 1 % natural image or piecewise smooth image
    K_min          = 5;     % minimium & maximum number of similar patches
    K_max          = 50;
    perc_rem       = 0.65;  % C_I in the paper
    perc_rem_last  = 0.9;   % C_I at the last iteration
    gamma = 0.6;            % gamma in the paper, key parameter controling the denoising effects
elseif strcmp(type, 'pws')
    K_min          = 5;
    K_max          = 60;
    perc_rem       = 0.7;
    perc_rem_last  = 1.15;
    gamma = 0;
else
    fprintf('Image type should either be natural (ntl) or piecewise smooth (pws).\n');
    return;
end
    
%% begin to iterates
for iter = 1 : iter_max
    % noise-dependent parameters
    if sigma_I <= 5 % (0,5]
        len_p = 7; th_K = 2e-3; step = 2;
        if (sigma_I <= sigma_th)
            perc_rem = perc_rem_last; % perform the last iteration
        end
    elseif sigma_I <= 10 % (5,10]
        len_p = 10; th_K = 3e-3; step = 2;
    elseif sigma_I <= 20 % (10,20]
        len_p = 12; th_K = 8e-3; step = 3;
    elseif sigma_I <= 30 % (20,30]
        len_p = 15; th_K = 2e-2; step = 3;
    else % natural image or piecewise smooth image
        if strcmp(type, 'ntl') == 1
            if sigma_I <= 50 % (30,50]
                len_p = 20; th_K = 3e-2; step = 4;
            elseif sigma_I <= 80 % (50,80]
                len_p = 20; th_K = 5e-2; step = 4;
            else % (80,\infty)
                len_p = 22; th_K = 8e-2; step = 6;
            end
        elseif strcmp(type, 'pws')
            if sigma_I <= 50 % (30,50]
                len_p = 20; th_K = 5e-2; step = 4;
            elseif sigma_I <= 80 % (50,80]
                len_p = 20; th_K = 6e-2; step = 4;
            else % (80,\infty)
                len_p = 22; th_K = 8e-2; step = 6;
            end
        end
    end

    [cnt_vert, cnt_hori] = deal(ceil((height - len_p) / step) + 1, ...
        ceil((width - len_p) / step) + 1);
    [f_1, f_2] = deal(reshape(repmat((1 : len_p)', [1 len_p]), [len_p^2 1]), ...
        reshape(repmat(1 : len_p, [len_p 1]), [len_p^2 1])); % spatial-related feature functions
    mask_diag = not(logical(eye(len_p^2)));
    eye_len = speye(len_p^2);
    
    %% construct the gradient filter matrices
    hx = [1 -1];
    hy = hx';
    grad_filt = zeros(2 * len_p * (len_p - 1), len_p^2);
    cnt = 0;
    for i = 1 : len_p % horizontal direction
        for j=1 : len_p-1
            cnt = cnt + 1;
            tmp=zeros(len_p);
            tmp(i, j : j + 1) = hx;
            grad_filt(cnt, :)=tmp(:)';
        end
    end
    for i=1 : len_p-1 % vertical direction
        for j=1 : len_p
            cnt = cnt + 1;
            tmp=zeros(len_p);
            tmp(i : i+1,j) = hy;
            grad_filt(cnt, :) = tmp(:)';
        end
    end

    %% preprocessing for patch clustering
    pat_den = zeros(len_p, len_p, cnt_vert * cnt_hori);
    th_dct=c_t_dct * sigma_I / 255;
    img_exp=reshape(transpose(im2col(single(img_int), [len_p len_p],'sliding')),[height- ...
        len_p + 1 width-len_p + 1 len_p^2]); % expand the noisy image
    parfor i = 1 : height-len_p + 1 % note the "parfor" loop here
        for j = 1 : width-len_p+1
            tmp = dct2(reshape(img_exp(i, j, :), [len_p len_p]));
            tmp(abs(tmp) < th_dct) = 0;
            img_exp(i, j, :) = reshape(tmp, [1 1 len_p^2]);
        end
    end
    idx_vert = zeros(ceil((height-len_p)/step)+1,1);
    idx_vert(1 : ceil((height - len_p + 1) / step)) = len_p : step : height;
    idx_vert(end) = height;
    idx_hori = zeros(ceil((width - len_p) / step) + 1, 1);
    idx_hori(1 : ceil((width - len_p + 1) / step)) = len_p : step : width;
    idx_hori(end) = width;
    [idx_vert, idx_hori] = deal(idx_vert - len_p + 1, idx_hori - len_p + 1);
    score_sim = zeros(cnt_vert * cnt_hori, 1);

    %% denoise each patch
    parfor idx_pat = 1 : cnt_vert * cnt_hori % note the "parfor" loop here
        
        %% initialization
        j = ceil(idx_pat / cnt_vert); % compute the coordinates
        i = idx_pat - (j - 1) * cnt_vert;
        [cur_vert, cur_hori] = deal(idx_vert(i), idx_hori(j));
        range_up = max(cur_vert - search_rad, 1);
        range_down = min(cur_vert + search_rad, height - len_p + 1);
        range_left = max(cur_hori - search_rad, 1);
        range_right = min(cur_hori + search_rad, width-len_p + 1);
        dif_cur = mean((img_exp(range_up : range_down, range_left : ...
            range_right, :) - repmat(img_exp(cur_vert, cur_hori, :), ...
            [range_down-range_up + 1 range_right - range_left + 1 1])).^2,3);
        [score_sim_tmp,idx_sim] = sort(dif_cur(:));
        pos_hori = ceil(idx_sim / (range_down - range_up + 1)); % compute the coordinates
        pos_vert = idx_sim - (pos_hori - 1) .* (range_down - range_up + 1);
        [pos_vert, pos_hori] = deal(pos_vert + range_up - 1, pos_hori + range_left - 1);
        K=max(min(sum(score_sim_tmp(:) < th_K), K_max), K_min);
        score_sim(idx_pat) = 1 / (1e-3 + mean(score_sim_tmp(1 : K)));
        
        %% perform patch clustering
        pat_sim = zeros(len_p^2, K);
        for cnt_pat = 1 : K % stack the similar patches
            pat_sim(:, cnt_pat) = reshape(img_int(pos_vert(cnt_pat) : ...
                pos_vert(cnt_pat) + len_p-1, pos_hori(cnt_pat) : ...
                pos_hori(cnt_pat) + len_p-1), [len_p^2 1]);
        end
        wght_pat = ones(1, K); wght_pat(1 : ceil(K / 3)) = 1;
        wght_pat = repmat(wght_pat / sum(wght_pat) * K, [len_p^2 1]);
        pat_sum = sum(pat_sim .* wght_pat,2);
        
        %% build the graph Laplacian L
        pat_noi = reshape(img_int(cur_vert : cur_vert + len_p - 1, cur_hori : ...
            cur_hori + len_p - 1), [len_p^2 1]); % current noisy patch
        sigma_g=sqrt(mean(mean((grad_filt * (pat_sim - repmat(pat_sum / K, ...
            [1 K]))).^2))) * c_sigma_I; % estimate the variance of the gradient
        sigma_til = sqrt(sigma_g^2 / (K + sigma_g^2 / sigma_p^2)); % variance of the posterior Gaussian
        F = [f_1 * sqrt(sigma_til^2 + alpha) f_2 * sqrt(sigma_til^2 + alpha) ...
            pat_sum / (K + sigma_g^2 / sigma_p^2)]; % all feature functions
        norm_node = sum(F .* F, 2);
        A = repmat(norm_node, [1 len_p^2]) + repmat(norm_node', [len_p^2 1])-2 * (F * F');
        A_sort = nth_element(A, edge_min + 1);
        r = max(A_sort(edge_min + 1, :)) + 1e-6;
        mask_tmp = triu(mask_diag & (A <= r)); % speed up with the upper-triangle characteristic
        A(not(mask_tmp)) = 0;
        eps = (max(grad_filt * pat_noi) + sigma_I / 255) * c_eps;
        A(mask_tmp) = exp(-A(mask_tmp) / (2 * eps^2));
        A = sparse(A);
        wght_rec = full((sum(A) + sum(A,2)').^(-gamma));
        wght_rec(wght_rec > 1e6) = 1e3; % deal with the case where the weight is almost zero
        A = (wght_rec'*wght_rec) .* A; % rectify the adjacency
        D = diag(sum(A) + sum(A, 2)'); % obtain the degree matrix
        L = (D - (A + A')) * (1e2 * len_p^2 / trace(D)); % obtain the graph Laplacian

        %% solve the constrained minimization problem
        deviation = len_p^2 * (sigma_I * perc_rem / 255)^2; % allowed deviation
        lambda_0 = lambda_0_const; lambda_1 = lambda_1_const; lambda_2 = (lambda_0 + lambda_1) / 2;
        pat_den_0 = (L + lambda_0*eye_len) \ (lambda_0 * pat_noi);
        dif_0 = norm(pat_den_0 - pat_noi)^2 - deviation;
        pat_den_1 = (L + lambda_1 * eye_len) \ (lambda_1 * pat_noi);
        dif_1 = norm(pat_den_1 - pat_noi)^2 - deviation;
        pat_den_2 = (L+lambda_2 * eye_len) \ (lambda_2 * pat_noi);
        dif_2 = norm(pat_den_2 - pat_noi)^2 - deviation;
        cnt_reg_iter = 0; % use the Muller's method to find lambda
        stop = false;
        while (abs(dif_2) > 1e-5) && (cnt_reg_iter < reg_iter_max) && (stop == false)
            cnt_reg_iter = cnt_reg_iter+1;
            h0 = lambda_0-lambda_2; h1 = lambda_1 - lambda_2; h2 = h0 * h1 * (h0-h1);
            a = (h1 * (dif_0-dif_2)-h0 * (dif_1-dif_2)) / h2;
            b = (h0^2 * (dif_1-dif_2)-h1^2 * (dif_0-dif_2)) / h2;
            lambda_0 = lambda_1; dif_0 = dif_1;
            lambda_1 = lambda_2; dif_1 = dif_2;
            lambda_2 = max(real(lambda_2 - 2 * dif_2 / (b + sign(b) * ...
                sqrt(b^2 - 4 * a * dif_2))), 1e-3);
            pat_den_2 = (L + lambda_2 * eye_len) \ (lambda_2 * pat_noi);
            dif_2 = norm(pat_den_2-pat_noi)^2 - deviation;
        end
        if isnan(dif_2) % NaN encountered
            pat_den_2 = pat_noi;
        end
        pat_den(:,:,idx_pat) = reshape(pat_den_2, [len_p len_p]);
    end
    
    %% assemble the denoised image
    img_den = zeros(height,width);
    confid_map = zeros(height,width);
%     score_sim = 1 ./ (score_sim + 1e-6);
    for idx_pat = 1 : cnt_vert * cnt_hori
        j = ceil(idx_pat / cnt_vert); % compute the coordinates
        i = idx_pat - (j - 1) * cnt_vert;
        [cur_vert, cur_hori] = deal(idx_vert(i), idx_hori(j));
        img_den(cur_vert : cur_vert + len_p-1, cur_hori : cur_hori + len_p - 1) = ...
            img_den(cur_vert : cur_vert + len_p-1, cur_hori : cur_hori + len_p - 1) + ...
            pat_den(:, :, idx_pat) * score_sim(idx_pat);
        confid_map(cur_vert : cur_vert + len_p-1, cur_hori : cur_hori + len_p-1) = ...
            confid_map(cur_vert : cur_vert + len_p-1, cur_hori : ...
            cur_hori+len_p-1) + score_sim(idx_pat);
    end
    img_int_prev = img_int;
    img_int = img_den ./ confid_map;
    if sigma_I > sigma_th;
        sigma_I = sigma_I - sqrt(mean((img_int(:) - img_int_prev(:)).^2)) * 255;
    else
        break;
    end
end
img_den = img_int;

end
